;;; File: "modules.scm"

;; Implements a simple module system on top of Gambit's ##namespace
;; special form.  It defines two special forms:
;;
;;   (##module <module-name> <language-name>
;;     (export <id> ...) ...
;;     (define-macro ...) ...
;;   )
;;
;;   (##require <module-name>)
;;
;; Currently, <module-name> must be a symbol naming the source file
;; without the extension and <language-name> must be the symbol
;; |gambit|.
;;
;; The ##module form must be the first expression in a file and
;; the ##require form can appear anywhere a definition can appear.

(##define-macro (##module-parse expand module-name language-name . rest)

  (define (syntax-error)
    (error "syntax error"))

  (define (valid-module-name? module-name)
    (symbol? module-name))

  (define (valid-language-name? language-name)
    (symbol? language-name))

  (define (str . rest)
    (apply string-append
           (map (lambda (s) (if (symbol? s) (symbol->string s) s))
                rest)))

  (define (sym . rest)
    (string->symbol (apply str rest)))

  (if (or (not (valid-module-name? module-name))
          (not (valid-language-name? language-name)))
      (syntax-error)
      (let loop ((lst rest)
                 (exports-rev '())
                 (defmacs-rev '()))
        (if (pair? lst)

            (let ((x (car lst)))
              (if (not (pair? x))
                  (syntax-error)
                  (case (car x)
                    ((export)
                     (loop (cdr lst)
                           (append (reverse (cdr x)) exports-rev)
                           defmacs-rev))
                    ((define-macro)
                     (let* ((pattern
                             (cadr x))
                            (id
                             (if (pair? pattern)
                                 (car pattern)
                                 pattern))
                            (lam
                             (if (pair? pattern)
                                 `(lambda ,(cdr pattern) ,@(cddr x))
                                 (caddr x))))
                       (loop (cdr lst)
                             exports-rev
                             (cons `(##define-macro ,id ,lam)
                                   defmacs-rev))))
                    (else
                     (syntax-error)))))

            (let* ((exports
                    (reverse exports-rev))
                   (defmacs
                     (map (lambda (x)
                            `(##define-macro
                              ,(let ((id (cadr x)))
                                 (if (memq id exports)
                                     (##exported-id module-name id)
                                     (##private-id module-name id)))
                              ,(caddr x)))
                          (reverse defmacs-rev)))
                   (expansion
                    `(,expand ,module-name
                              ,language-name
                              ,(if (null? exports)
                                   `()
                                   `((##namespace (,(str module-name "#")
                                                   ,@exports))))
                              ,defmacs)))
;;              (pp expansion)
              expansion)))))

(##define-macro (##module module-name language-name . rest)
  `(##module-parse ##module-expand ,module-name ,language-name ,@rest))

(##define-macro (##module-expand module-name language-name exports defmacs)

  (define (str . rest)
    (apply string-append
           (map (lambda (s) (if (symbol? s) (symbol->string s) s))
                rest)))

  (define (sym . rest)
    (string->symbol (apply str rest)))

  (let ((expansion
         `(begin
            (##namespace (,(str module-name "##")))
            (##include ,(str "~~lib/" language-name "#.scm"))
            ,@defmacs
            ,@exports)))
;;    (pp expansion)
    expansion))

(##define-macro (##require module-name)

  (define (module-form module-name)
    (let* ((source-file (##module-name->module-source-file module-name))
           (x (with-input-from-file source-file read)))
      (if (and (pair? x) (eq? (car x) '##module))
          x
          #f)))

  (let ((mf (module-form module-name)))
    (if mf
        `(##module-parse ##require-expand ,@(cdr mf))
        `(begin))))

(##define-macro (##require-expand module-name language-name exports defmacs)
  (let ((expansion
         `(begin
            ,@defmacs
            ,@exports
            (##load-module-once ',module-name) ;; should really be at top of file
            )))
;;    (pp expansion)
    expansion))

(define (##exported-id module-name id)
  (string->symbol
   (string-append (symbol->string module-name)
                  "#"
                  (symbol->string id))))

(define (##private-id module-name id)
  (string->symbol
   (string-append (symbol->string module-name)
                  "##"
                  (symbol->string id))))

(define ##loaded-modules (make-table))

(define (##load-module-once module-name)
  (let ((x (table-ref ##loaded-modules module-name #f)))
    (if (not x)
        (let ((filename (##module-name->module-file module-name)))
          (table-set! ##loaded-modules module-name #t)
          (load filename)))))

(define (##module-name->module-source-file module-name)
  (string-append (##module-name->module-file module-name)
                 ".scm"))

(define (##module-name->module-file module-name)
  (symbol->string module-name))
